# coding=utf-8
"""Utility functions for calculating the Humidex."""
from __future__ import division

import math


def humidex(ta, tdp):
    """Calculate Humidex from air temperature and the Dew Point.

    The Humidex is a Canadian innovation first used in 1965.
    It combines the temperature and humidity into one number to reflect the
    perceived temperature.
    Because it takes into account the two most important
    factors that affect summer comfort, it can be a better
    measure of how stifling the air feels than either temperature or
    humidity alone. [1]

    Air temperatures below 20c will give a generally meaningless result as the Humidex
    only describes perceived heat.

    The Humidex is a "nominally dimensionless quantity" but is generally
    recognized by the public as equivalent to the degree Celsius [2]

    Note:
        [1] Environment Canada (October 2017). "Warm Season Weather Hazards".
            https://www.canada.ca/en/environment-climate-change/services/seasonal\
-weather-hazards/warm-season-weather-hazards.html#toc7

        [2] https://en.wikipedia.org/wiki/Humidex

    Args:
        ta: Air temperature [C]
        tdp: The Dew Point [C]

    Returns:
        float -- Humidex
    """

    dew_point_k = tdp + 273.15  # celsius to kelvin

    e = 6.11 * math.exp(5417.7530 * ((1 / 273.15) - (1 / dew_point_k)))
    h = 0.5555 * (e - 10.0)

    humidex_value = float(ta + h)
    return humidex_value


def humidex_degree_of_comfort(humidex):
    """Get the degree of comfort associated with a given Humidex value.

    Degrees of comfort are provided by the Government of Canada and are indicated
    here with the following integer values:

    * 0 = No discomfort (Humidex of 19 and below)
    * 1 = Little discomfort (Humidex between 20 - 29)
    * 2 = Some discomfort (Humidex between 30 - 39)
    * 3 = Great discomfort; avoid exertion (Humidex between 40 - 45)
    * 4 = Dangerous; heat stroke possible (Humidex of 45 and above)

    See: https://www.canada.ca/en/environment-climate-change/services/seasonal-\
weather-hazards/warm-season-weather-hazards.html#toc7

    Args:
        humidex: Humidex

    Returns:
        int -- Degree of Comfort
    """

    if humidex < 20.0:
        return 0
    elif humidex < 30.0:
        return 1
    elif humidex < 40.0:
        return 2
    elif humidex < 46.0:
        return 3
    return 4



# coding=utf-8
"""Object for plotting a Adaptive Comfort Chart."""
from __future__ import division

from ladybug_geometry.geometry2d import Point2D, Vector2D, LineSegment2D, \
    Polyline2D, Polygon2D, Mesh2D
from ladybug_geometry.geometry3d import Point3D, Vector3D

from ladybug.datacollection import DailyCollection, MonthlyCollection
from ladybug.graphic import GraphicContainer
from ladybug.datatype.time import Time
from ladybug.datatype.temperaturedelta import TemperatureDelta
from ladybug.datatype.temperature import Temperature, OperativeTemperature

from ..adaptive import neutral_temperature_ashrae55, neutral_temperature_en15251, \
    neutral_temperature_conditioned_function, cooling_effect_ashrae55, \
    cooling_effect_en15251, t_operative
from ..collection.base import BaseCollection
from ..collection.adaptive import Adaptive


class AdaptiveChart(object):
    """Adaptive comfort DataCollection object.

    Args:
        outdoor_temperature: Either one of the following inputs are acceptable:

            * A Data Collection of prevailing outdoor temperature values in C.
              Such a Data Collection must align with the operative_temperature
              input and bear the PrevailingOutdoorTemperature data type in
              its header.
            * A single prevailing outdoor temperature value in C to be used
              for all of the operative_temperature inputs below.
            * A Data Collection of actual outdoor temperatures recorded over
              the entire year. This Data Collection must be continuous and
              must either be an Hourly Collection or Daily Collection. In the event
              that the input comfort_parameter has a prevailing_temperature_method
              of 'Monthly', Monthly collections are also acceptable here. Note
              that, because an annual input is required, this input collection
              does not have to align with the operative_temperature input.

        operative_temperature: Data Collection of operative temperature (To)
            values in degrees Celsius.
        air_speed: A number for the air speed values in m/s. If None, a low air
            speed of 0.1 m/s wil be used. (Default: None).
        comfort_parameter: Optional AdaptiveParameter object to specify parameters
            under which conditions are considered acceptable. If None, default will
            assume ASHRAE-55 criteria.
        legend_parameters: An optional LegendParameter object to change the display
            of the AdaptiveChart. (Default: None).
        base_point: A Point2D to be used as a starting point to generate the geometry
            of the plot. (Default: (0, 0)).
        x_dim: A number to set the X dimension of each degree of temperature on the
            chart. (Default: 1).
        y_dim: A number to set the Y dimension of each degree of temperature on
            the chart. (Default: 1).
            Note that most maximum humidity ratios are around 0.03. (Default: 1500).
        min_prevailing: An integer for the minimum prevailing temperature on the
            chart in degrees. This should be celsius if use_ip is False and fahrenheit
            if use_ip is True. (Default: 10; suitable for celsius).
        max_prevailing: An integer for the maximum prevailing temperature on the
            chart in degrees. This should be celsius if use_ip is False and fahrenheit
            if use_ip is True. (Default: 33; suitable for celsius).
        min_operative: An integer for the minimum indoor operative temperature on the
            chart in degrees. This should be celsius if use_ip is False and fahrenheit
            if use_ip is True. (Default: 14; suitable for celsius).
        max_operative: An integer for the maximum indoor operative temperature on the
            chart in degrees. This should be celsius if use_ip is False and fahrenheit
            if use_ip is True. (Default: 40; suitable for celsius).
        use_ip: Boolean to note whether temperature values should be plotted in
            Fahrenheit instead of Celsius. (Default: False).

    Properties:
        * collection
        * prevailing_outdoor_temperature
        * operative_temperature
        * air_speed
        * comfort_parameter
        * legend_parameters
        * base_point
        * x_dim
        * y_dim
        * min_prevailing
        * max_prevailing
        * min_operative
        * max_operative
        * use_ip
        * chart_border
        * prevailing_labels
        * prevailing_label_points
        * prevailing_lines
        * operative_labels
        * operative_label_points
        * operative_lines
        * neutral_temperature
        * degrees_from_neutral
        * neutral_polyline
        * comfort_polygon
        * is_comfortable
        * thermal_condition
        * percent_comfortable
        * percent_uncomfortable
        * percent_neutral
        * percent_hot
        * percent_cold
        * title_text
        * title_location
        * x_axis_text
        * x_axis_location
        * y_axis_text
        * y_axis_location
        * data_points
        * time_matrix
        * hour_values
        * colored_mesh
        * legend
        * container
    """
    __slots__ = (
        '_collection', '_prevailing_outdoor_temperature', '_operative_temperature',
        '_neutral_temperature', '_degrees_from_neutral',
        '_base_point', '_x_dim', '_y_dim',
        '_min_prevailing', '_max_prevailing', '_min_operative', '_max_operative',
        '_use_ip', '_tp_category', '_to_category', '_prevail_range', '_op_range',
        '_x_range', '_y_range', '_time_multiplier',
        '_time_matrix', '_hour_values', '_remove_pattern', '_container',
        '_chart_border', '_data_points', '_colored_mesh'
    )
    TEMP_TYPE = Temperature()
    DT_TYPE = TemperatureDelta()

    def __init__(
        self, outdoor_temperature, operative_temperature, air_speed=None,
        comfort_parameter=None, legend_parameters=None, base_point=Point2D(),
        x_dim=1, y_dim=1, min_prevailing=10, max_prevailing=33,
        min_operative=14, max_operative=40, use_ip=False
    ):
        # check inputs that determine other inputs
        assert air_speed is None or isinstance(air_speed, (float, int)), \
            'Expected number or None for air_speed. Got {}.'.format(type(air_speed))
        self._use_ip = bool(use_ip)

        # build an adaptive comfort collection from the data
        self._collection = Adaptive(outdoor_temperature, operative_temperature,
                                    air_speed, comfort_parameter)

        # ensue all temperatures are in the correct units
        if self._use_ip:  # convert everything to Fahrenheit
            self._prevailing_outdoor_temperature = \
                self._collection.prevailing_outdoor_temperature.to_ip()
            self._operative_temperature = \
                self._collection.operative_temperature.to_ip()
            self._neutral_temperature = \
                self._collection.neutral_temperature.to_ip()
            self._degrees_from_neutral = \
                self._collection.degrees_from_neutral.to_ip()
        else:
            self._prevailing_outdoor_temperature = \
                self._collection.prevailing_outdoor_temperature
            self._operative_temperature = \
                self._collection.operative_temperature
            self._neutral_temperature = self._collection.neutral_temperature
            self._degrees_from_neutral = self._collection.degrees_from_neutral

        # extract the timestep from the data collections
        if isinstance(self._operative_temperature, MonthlyCollection):
            self._time_multiplier = 30 * 24
        elif isinstance(self._operative_temperature, DailyCollection):
            self._time_multiplier = 24
        else:  # it's an hourly or sub-hourly collection
            self._time_multiplier = \
                1 / self._operative_temperature.header.analysis_period.timestep

        # assign the inputs as properties of the chart
        assert isinstance(base_point, Point2D), 'Expected Point2D for ' \
            'PsychrometricChart base point. Got {}.'.format(type(base_point))
        self._base_point = base_point
        self._x_dim = self._check_number(x_dim, 'x_dim')
        self._y_dim = self._check_number(y_dim, 'y_dim')
        assert max_prevailing - min_prevailing >= 10, 'Adaptive chart ' \
            'max_prevailing and min_prevailing difference must be at least 10.'
        assert max_operative - min_operative >= 10, 'Adaptive chart ' \
            'max_operative and min_operative difference must be at least 10.'
        self._max_prevailing = int(max_prevailing)
        self._min_prevailing = int(min_prevailing)
        self._max_operative = int(max_operative)
        self._min_operative = int(min_operative)
        if self._use_ip:
            assert self._max_prevailing > 50, 'max_prevailing must be greater than ' \
                '50F. Got {}.'.format(self._max_prevailing)
            assert self._min_prevailing < 86, 'min_prevailing must be less than ' \
                '86F. Got {}.'.format(self._min_prevailing)
        else:
            assert self._max_prevailing > 10, 'max_prevailing must be greater than ' \
                '10C. Got {}.'.format(self._max_prevailing)
            assert self._min_prevailing < 30, 'min_prevailing must be less than ' \
                '30C. Got {}.'.format(self._min_prevailing)

        # create the graphic container
        if self._use_ip:  # categorize based on every 1.66 fahrenheit
            self._tp_category, self._to_category = [], []
            current_t, max_t = self._min_prevailing, self._max_prevailing + 1.75
            while current_t < max_t:
                current_t += (5 / 3)
                self._tp_category.append(current_t)
            current_t, max_t = self._min_operative, self._max_operative + 1.75
            while current_t < max_t:
                current_t += (5 / 3)
                self._to_category.append(current_t)
        else:  # categorize based on every degree celsius
            self._tp_category = list(range(self._min_prevailing + 1,
                                           self._max_prevailing + 1))
            self._to_category = list(range(self._min_operative + 1,
                                           self._max_operative + 1))
        self._time_matrix, self._hour_values, self._remove_pattern = \
            self._compute_hour_values()
        assert len(self._hour_values) > 0, \
            'No data was found to lie on the adaptive chart.'
        max_x = base_point.x + (self._max_prevailing - self._min_prevailing + 1) \
            * self._x_dim
        max_y = base_point.y + (self._max_operative - self._min_operative + 1) \
            * self._y_dim
        max_pt = Point3D(max_x, max_y, 0)
        min_pt = Point3D(base_point.x, base_point.y, 0)
        self._container = GraphicContainer(
            self._hour_values, min_pt, max_pt, legend_parameters, Time(), 'hr')
        self._process_legend_default(self._container.legend_parameters)

        # create global attributes used by several of the geometry properties
        self._prevail_range = list(range(self._min_prevailing, self._max_prevailing, 2)) \
            + [self._max_prevailing]
        self._x_range = [self.tp_x_value(t) for t in self._prevail_range]
        self._op_range = list(range(self._min_operative, self._max_operative, 2)) \
            + [self._max_operative]
        self._y_range = [self.to_y_value(t) for t in self._op_range]
        if use_ip:  # ensure that _temp_range is always in celsius
            self._prevail_range = self.TEMP_TYPE.to_unit(self._prevail_range, 'C', 'F')
            self._op_range = self.TEMP_TYPE.to_unit(self._op_range, 'C', 'F')

        # set null values for properties that are optional
        self._chart_border = None
        self._data_points = None
        self._colored_mesh = None

    @classmethod
    def from_air_and_rad_temp(
        cls, outdoor_temperature, air_temperature, rad_temperature=None, air_speed=None,
        comfort_parameter=None, legend_parameters=None, base_point=Point2D(),
        x_dim=1, y_dim=1, min_prevailing=10, max_prevailing=33,
        min_operative=14, max_operative=40, use_ip=False
    ):
        """Initialize an AdaptiveChart from air and radiant temperature."""
        if rad_temperature is None:
            to = air_temperature
        else:
            to = BaseCollection.compute_function_aligned(
                t_operative, [air_temperature, rad_temperature],
                OperativeTemperature(), 'C')
        return cls(
            outdoor_temperature, to, air_speed, comfort_parameter,
            legend_parameters, base_point, x_dim, y_dim,
            min_prevailing, max_prevailing, min_operative, max_operative, use_ip
        )

    @property
    def prevailing_outdoor_temperature(self):
        """Data Collection of prevailing outdoor temperature.

        This will be in in degrees C if use_ip is False and degrees F if use_ip is True.
        """
        return self._prevailing_outdoor_temperature

    @property
    def operative_temperature(self):
        """Data Collection of indoor operative temperature.

        This will be in in degrees C if use_ip is False and degrees F if use_ip is True.
        """
        return self._operative_temperature

    @property
    def air_speed(self):
        """Value for air speed in m/s."""
        return self._collection.air_speed[0]

    @property
    def comfort_parameter(self):
        """Adaptive comfort parameters that are assigned to this object."""
        return self._collection.comfort_parameter

    @property
    def legend_parameters(self):
        """The legend parameters customizing this adaptive chart."""
        return self._container.legend_parameters

    @property
    def base_point(self):
        """Point3D for the base point of this adaptive chart."""
        return self._base_point

    @property
    def x_dim(self):
        """The X dimension for each degree of prevailing temperature on the chart."""
        return self._x_dim

    @property
    def y_dim(self):
        """The Y dimension for each degree of operative temperature on the chart."""
        return self._y_dim

    @property
    def min_prevailing(self):
        """An integer for the minimum prevailing outdoor temperature on the chart.

        Will be in celsius if use_ip is False and fahrenheit if use_ip is True.
        """
        return self._min_prevailing

    @property
    def max_prevailing(self):
        """An integer for the maximum prevailing outdoor temperature on the chart.

        Will be in celsius if use_ip is False and fahrenheit if use_ip is True.
        """
        return self._max_prevailing

    @property
    def min_operative(self):
        """An integer for the minimum indoor operative temperature on the chart.

        Will be in celsius if use_ip is False and fahrenheit if use_ip is True.
        """
        return self._min_operative

    @property
    def max_operative(self):
        """An integer for the maximum indoor operative temperature on the chart.

        Will be in celsius if use_ip is False and fahrenheit if use_ip is True.
        """
        return self._max_operative

    @property
    def use_ip(self):
        """Boolean for whether temperature should be in Fahrenheit or Celsius."""
        return self._use_ip

    @property
    def chart_border(self):
        """Get a Polygon2D for the border of the chart."""
        if self._chart_border is None:
            self._chart_border = self._compute_border()
        return self._chart_border

    @property
    def prevailing_labels(self):
        """Get a tuple of text for the prevailing temperature labels on the chart."""
        if self.use_ip:
            temp_range = tuple(range(self._min_prevailing, self._max_prevailing, 2)) \
                + (self._max_prevailing,)
            return tuple(str(val) for val in temp_range)
        return tuple(str(val) for val in self._prevail_range)

    @property
    def prevailing_label_points(self):
        """Get a tuple of Point2Ds for the prevailing temperature labels on the chart."""
        y_val = self._base_point.y - self.legend_parameters.text_height * 0.5
        return tuple(Point2D(x_val, y_val) for x_val in self._x_range)

    @property
    def prevailing_lines(self):
        """Get a tuple of LineSegment2Ds for the prevailing temperature lines."""
        t_lines = []  # create the array of line segments
        y_vec = Vector2D(0, self._y_range[-1] - self._y_range[0])
        for x_val in self._x_range:
            l_seg = LineSegment2D(Point2D(x_val, self._base_point.y), y_vec)
            t_lines.append(l_seg)
        return t_lines

    @property
    def operative_labels(self):
        """Get a tuple of text for the operative temperature labels on the chart."""
        if self.use_ip:
            temp_range = tuple(range(self._min_operative, self._max_operative, 2)) \
                + (self._max_operative,)
            return tuple(str(val) for val in temp_range)
        return tuple(str(val) for val in self._op_range)

    @property
    def operative_label_points(self):
        """Get a tuple of Point2Ds for the operative temperature labels on the chart."""
        x_val = self._base_point.x - self.legend_parameters.text_height * 2.5
        return tuple(Point2D(x_val, y_val) for y_val in self._y_range)

    @property
    def operative_lines(self):
        """Get a tuple of LineSegment2Ds for the operative temperature lines."""
        t_lines = []  # create the array of line segments
        x_vec = Vector2D(self._x_range[-1] - self._x_range[0], 0)
        for y_val in self._y_range:
            l_seg = LineSegment2D(Point2D(self._base_point.x, y_val), x_vec)
            t_lines.append(l_seg)
        return t_lines

    @property
    def neutral_polyline(self):
        """Get a LineSegment2D or Polyline2D noting the neutral temperature on the chart.
        """
        # get properties that are used to compute the neutral temperature
        tp_c_min, tp_c_max = self._prevail_range[0], self._prevail_range[-1]
        pl_pts = []
        if self.comfort_parameter.conditioning != 0:
            neutral_func = neutral_temperature_conditioned_function(
                self.comfort_parameter.conditioning, self.comfort_parameter.standard
            )
        elif self.comfort_parameter.ashrae_or_en:
            neutral_func = neutral_temperature_ashrae55
        else:
            neutral_func =neutral_temperature_en15251

        # get the beginning points
        x_val1 = self._x_range[0]
        if tp_c_min < 10:
            n_temp = neutral_func(10)
            y_val = self.to_y_value(n_temp) if not self.use_ip else \
                self.to_y_value(self.TEMP_TYPE.to_unit([n_temp], 'F', 'C')[0])
            pl_pts.append(Point2D(x_val1, y_val))
            x_val2 = self.tp_x_value(50) if self.use_ip else self.tp_x_value(10)
            pl_pts.append(Point2D(x_val2, y_val))
        else:
            n_temp = neutral_func(tp_c_min)
            y_val = self.to_y_value(n_temp) if not self.use_ip else \
                self.to_y_value(self.TEMP_TYPE.to_unit([n_temp], 'F', 'C')[0])
            pl_pts.append(Point2D(x_val1, y_val))
        # get the ending points
        x_val_end = self._x_range[-1]
        if self.comfort_parameter.ashrae_or_en:
            if tp_c_max > 33.5:
                n_temp = neutral_func(33.5)
                y_val = self.to_y_value(n_temp) if not self.use_ip else \
                    self.to_y_value(self.TEMP_TYPE.to_unit([n_temp], 'F', 'C')[0])
                x_vali = self.tp_x_value(92.3) if self.use_ip else self.tp_x_value(33.5)
                pl_pts.append(Point2D(x_vali, y_val))
                pl_pts.append(Point2D(x_val_end, y_val))
            else:
                n_temp = neutral_func(tp_c_max)
                y_val = self.to_y_value(n_temp) if not self.use_ip else \
                    self.to_y_value(self.TEMP_TYPE.to_unit([n_temp], 'F', 'C')[0])
                pl_pts.append(Point2D(x_val_end, y_val))
        else:
            if tp_c_max > 30:
                n_temp = neutral_func(30)
                y_val = self.to_y_value(n_temp) if not self.use_ip else \
                    self.to_y_value(self.TEMP_TYPE.to_unit([n_temp], 'F', 'C')[0])
                x_vali = self.tp_x_value(86) if self.use_ip else self.tp_x_value(30)
                pl_pts.append(Point2D(x_vali, y_val))
                pl_pts.append(Point2D(x_val_end, y_val))
            else:
                n_temp = neutral_func(tp_c_max)
                y_val = self.to_y_value(n_temp) if not self.use_ip else \
                    self.to_y_value(self.TEMP_TYPE.to_unit([n_temp], 'F', 'C')[0])
                pl_pts.append(Point2D(x_val_end, y_val))

        # return the neutral line
        return Polyline2D(pl_pts) if len(pl_pts) > 2 else \
            LineSegment2D.from_end_points(pl_pts[0], pl_pts[1])

    @property
    def comfort_polygon(self):
        """Get a Polygon2D for the comfort range on the chart."""
        # start off with the neutral polyline and move it based on the offset
        neutral_line = self.neutral_polyline
        offset_t = self.comfort_parameter.neutral_offset if not self.use_ip else \
            self.DT_TYPE.to_unit([self.comfort_parameter.neutral_offset], 'dF', 'dC')[0]
        offset_dist = self.y_dim * offset_t
        lower_line = neutral_line.move(Vector2D(0, -offset_dist))
        upper_line = neutral_line.move(Vector2D(0, offset_dist))

        # trim the bottom of the polygon if there is a cold_prevail_temp_limit
        if self.comfort_parameter.cold_prevail_temp_limit > 10:
            limit_tc = self.comfort_parameter.cold_prevail_temp_limit
            limit_t = limit_tc if not self.use_ip else \
                self.TEMP_TYPE.to_unit([limit_tc], 'F', 'C')[0]
            limit_x = self.tp_x_value(limit_t)
            int_lin = LineSegment2D.from_end_points(Point2D(limit_x, self._y_range[0]),
                                                    Point2D(limit_x, self._y_range[-1]))
            i_pts = lower_line.intersect_line_ray(int_lin)
            if i_pts is not None and (len(i_pts) == 1 or isinstance(i_pts, Point2D)):
                int_pt = i_pts if isinstance(i_pts, Point2D) else i_pts[0]
                new_low_pts, int_passed = [], False
                for pt in lower_line.vertices:
                    if pt.x < int_pt.x:
                        new_low_pts.append(Point2D(pt.x, int_pt.y))
                    elif not int_passed:
                        new_low_pts.append(int_pt)
                        new_low_pts.append(pt)
                        int_passed = True
                    else:
                        new_low_pts.append(pt)
                lower_line = Polyline2D(new_low_pts) if len(new_low_pts) > 2 else \
                    LineSegment2D.from_end_points(new_low_pts[0], new_low_pts[1])

        # determine if there is a cooling effect
        if self.comfort_parameter.discrete_or_continuous_air_speed is True:
            cooling_func = cooling_effect_ashrae55
        else:
            cooling_func = cooling_effect_en15251
        ce = cooling_func(self.air_speed, self._prevail_range[-1])
        if ce == 0:  # we can build the polygon from upper/lower lines
            return Polygon2D(lower_line.vertices + tuple(reversed(upper_line.vertices)))

        # adjust the upper line to account for the cooling effect
        ce_t = ce if not self.use_ip else self.DT_TYPE.to_unit([ce], 'dF', 'dC')[0]
        ce_dist = self.y_dim * ce_t
        ce_vec = Vector2D(0, ce_dist)
        switch_tc = 12 if self.comfort_parameter.ashrae_or_en else 12.73
        switch_t = switch_tc if not self.use_ip else \
            self.TEMP_TYPE.to_unit([switch_tc], 'F', 'C')[0]
        switch_x = self.tp_x_value(switch_t)
        if upper_line.vertices[0].x >= switch_x:
            new_up_pts = [pt.move(ce_vec) for pt in upper_line.vertices]
        else:
            new_up_pts, switch_occurred = [], False
            for i, pt in enumerate(upper_line.vertices):
                if pt.x <= switch_x:
                    new_up_pts.append(pt)
                else:
                    if switch_occurred:
                        new_up_pts.append(pt.move(ce_vec))
                    else:
                        int_line1 = LineSegment2D.from_end_points(
                            Point2D(switch_x, self._y_range[0]),
                            Point2D(switch_x, self._y_range[-1]))
                        int_line2 = LineSegment2D.from_end_points(
                            upper_line.vertices[i - 1], pt)
                        int_pt = int_line1.intersect_line_ray(int_line2)
                        new_up_pts.append(int_pt)
                        new_up_pts.append(int_pt.move(ce_vec))
                        new_up_pts.append(pt.move(ce_vec))
                        switch_occurred = True
        return Polygon2D(lower_line.vertices + tuple(reversed(new_up_pts)))

    @property
    def neutral_temperature(self):
        """Data Collection of the desired neutral temperature in degrees C."""
        return self._neutral_temperature

    @property
    def degrees_from_neutral(self):
        """Data Collection of the degrees from desired neutral temperature in C."""
        return self._degrees_from_neutral

    @property
    def is_comfortable(self):
        """Data Collection of integers noting whether the input conditions are
        acceptable according to the assigned comfort_parameter.

        Values are one of the following:
        * 0 = uncomfortable
        * 1 = comfortable
        """
        return self._collection.is_comfortable

    @property
    def thermal_condition(self):
        """Data Collection of integers noting the thermal status of a subject
        according to the assigned comfort_parameter.

        Values are one of the following:
        * -1 = cold
        * 0 = neutral
        * +1 = hot
        """
        return self._collection.thermal_condition

    @property
    def percent_comfortable(self):
        """The percent of time comfortable given by the assigned comfort_parameter."""
        return self._collection.percent_comfortable

    @property
    def percent_uncomfortable(self):
        """The percent of time uncomfortable given by the assigned comfort_parameter."""
        return self._collection.percent_uncomfortable

    @property
    def percent_neutral(self):
        """The percent of time that the thermal_condition is neutral."""
        self._collection.percent_neutral

    @property
    def percent_cold(self):
        """The percent of time that the thermal_condition is cold."""
        self._collection.percent_cold

    @property
    def percent_hot(self):
        """The percent of time that the thermal_condition is hot."""
        self._collection.percent_hot

    @property
    def title_text(self):
        """Get text for the title of the chart."""
        title_items = ['Adaptive Chart', 'Time [hr]']
        extra_data = self.operative_temperature.header.metadata.items()
        if len(extra_data) == 0:
            extra_data = self.prevailing_outdoor_temperature.header.metadata.items()
        return '\n'.join(title_items + ['{}: {}'.format(k, v) for k, v in extra_data])

    @property
    def title_location(self):
        """Get a Point2D for the title of the chart."""
        origin = self.container.lower_title_location.o.move(
            Vector3D(0, -self.legend_parameters.text_height * 4))
        return Point2D(origin.x, origin.y)

    @property
    def x_axis_text(self):
        """Get text for the X-axis label of the chart."""
        unit = 'C' if not self.use_ip else 'F'
        return 'Prevailing Outdoor Temperature [{}]'.format(unit) \
            if self.comfort_parameter.avg_month_or_running_mean else \
            'Outdoor Running Mean Temperature [{}]'.format(unit)

    @property
    def x_axis_location(self):
        """Get a Point2D for the X-axis label of the chart."""
        y_val = self._base_point.y - self.legend_parameters.text_height * 2.5
        x_val = (self._x_range[0] + self._x_range[-1]) / 2
        return Point2D(x_val, y_val)

    @property
    def y_axis_text(self):
        """Get text for the Y-axis label of the chart."""
        unit = 'C' if not self.use_ip else 'F'
        if 'type' in self.operative_temperature.header.metadata:
            return '{} [{}]'.format(
                self.operative_temperature.header.metadata['type'], unit)
        return '{} [{}]'.format(self.operative_temperature.header.data_type, unit)

    @property
    def y_axis_location(self):
        """Get a Point2D for the Y-axis label of the chart."""
        x_val = self._base_point.x - self.legend_parameters.text_height * 5.5
        y_val = (self._y_range[0] + self._y_range[-1]) / 2
        return Point2D(x_val, y_val)

    @property
    def data_points(self):
        """Get a tuple of Point2Ds for each of the temperature values."""
        if self._data_points is None:
            zip_o = zip(self.prevailing_outdoor_temperature, self.operative_temperature)
            self._data_points = tuple(
                Point2D(self.tp_x_value(tp), self.to_y_value(to))
                for tp, to in zip_o
            )
        return self._data_points

    @property
    def time_matrix(self):
        """Get a tuple of of tuples where each sub-tuple is a row of the mesh.

        Each value in the resulting matrix corresponds to the number of prevailing/
        operative temperature points in a given cell of the mesh.
        """
        return tuple(tuple(row) for row in self._time_matrix)

    @property
    def hour_values(self):
        """Get a tuple for the number of hours associated with each colored_mesh face."""
        return self._hour_values

    @property
    def colored_mesh(self):
        """Get a colored mesh for the number of hours for each part of the chart."""
        if self._colored_mesh is None:
            self._colored_mesh = self._generate_mesh()
        return self._colored_mesh

    @property
    def legend(self):
        """The legend assigned to this graphic."""
        return self._container._legend

    @property
    def container(self):
        """Get the GraphicContainer for the colored mesh."""
        return self._container

    def data_mesh(self, data_collection, legend_parameters=None):
        """Get a colored mesh for a data_collection aligned with the chart's data.

        Args:
            data_collection: A data collection that is aligned with the prevailing
                and operative temperature values of the chart.
            legend_parameters: Optional legend parameters to customize the legend
                and look of the resulting mesh.

        Returns:
            A tuple with two values.

            -   mesh: A colored Mesh2D similar to the chart's colored_mesh property
                but where each face is colored with the average value of the input
                data_collection.

            -   container: A GraphicContainer object for the mesh, which possesses
                a legend that corresponds to the mesh.
        """
        # check to be sure the data collection aligns
        data_vals = data_collection.values
        _tp_values = self.prevailing_outdoor_temperature.values
        _to_values = self.operative_temperature.values
        assert len(data_vals) == len(self.operative_temperature.values), \
            'Number of data collection values ' \
            'must match those of the prevailing and operative temperature.'

        # create a matrix with a tally of the hours for all the data
        base_mtx = [[[] for val in self._tp_category] for rh in self._to_category]
        for tp, to, val in zip(_tp_values, _to_values, data_vals):
            if tp < self._min_prevailing or tp > self._max_prevailing:
                continue  # temperature value does not currently fit on the chart
            if to < self._min_operative or to > self._max_operative:
                continue  # temperature value does not currently fit on the chart
            for y, to_cat in enumerate(self._to_category):
                if to < to_cat:
                    break
            for x, tp_cat in enumerate(self._tp_category):
                if tp < tp_cat:
                    break
            base_mtx[y][x].append(val)

        # compute average values
        avg_values = [sum(val_list) / len(val_list) for tp_l in base_mtx
                      for val_list in tp_l if len(val_list) != 0]

        # create the colored mesh and graphic container
        base_contain = self.container
        container = GraphicContainer(
            avg_values, base_contain.min_point, base_contain.max_point,
            legend_parameters, data_collection.header.data_type,
            data_collection.header.unit)
        self._process_legend_default(container.legend_parameters)
        mesh = self.colored_mesh.duplicate()  # start with hour mesh as a base
        mesh.colors = container.value_colors
        return mesh, container

    def plot_point(self, prevailing, operative):
        """Get a Point2D for a given prevailing and operative temperature on the chart.

        Args:
            prevailing: A prevailing temperature value, which should be in Celsius
                if use_ip is False and Fahrenheit is use_ip is True.
            operative: An operative temperature value, which should be in Celsius
                if use_ip is False and Fahrenheit is use_ip is True.
        """
        return Point2D(self.tp_x_value(prevailing), self.to_y_value(operative))

    def to_y_value(self, temperature):
        """Get the Y-coordinate for a certain operative temperature on the chart.

        Args:
            temperature: A temperature value, which should be in Celsius if use_ip
                is False and Fahrenheit is use_ip is True.
        """
        return self.base_point.y + self._y_dim * (temperature - self._min_operative)

    def tp_x_value(self, temperature):
        """Get the X-coordinate for a certain prevailing temperature on the chart.

        Args:
            temperature: A temperature value, which should be in Celsius if use_ip
                is False and Fahrenheit is use_ip is True.
        """
        return self._base_point.x + self._x_dim * (temperature - self._min_prevailing)

    def _compute_hour_values(self):
        """Compute the matrix of binned time values based on the chart inputs.

        Returns:
            A tuple with three values.

            -   base_mtx: A full matrix with counts of values for each degree
                temperature and 5% RH of the chart.

            -   mesh_values: A list of numbers for the values of the mesh.

            -   remove_pattern: A list of booleans for which faces of the full mesh
                should be removed.
        """
        # create a matrix with a tally of the hours for all the data
        base_mtx = [[0 for tp in self._tp_category] for to in self._to_category]
        zip_obj = zip(self.prevailing_outdoor_temperature, self.operative_temperature)
        for tp, to in zip_obj:
            if tp < self._min_prevailing or tp > self._max_prevailing:
                continue  # temperature value does not currently fit on the chart
            if to < self._min_operative or to > self._max_operative:
                continue  # temperature value does not currently fit on the chart
            for y, to_cat in enumerate(self._to_category):
                if to < to_cat:
                    break
            for x, tp_cat in enumerate(self._tp_category):
                if tp < tp_cat:
                    break
            base_mtx[y][x] += 1

        # flatten the matrix and create a pattern to remove faces
        flat_values = [tc * self._time_multiplier for to_l in base_mtx for tc in to_l]
        remove_pattern = [val != 0 for val in flat_values]
        mesh_values = tuple(val for val in flat_values if val != 0)
        return base_mtx, mesh_values, remove_pattern

    def _generate_mesh(self):
        """Get the colored mesh from this object's hour values."""
        # global properties used in the generation of the mesh
        t_per_row = [self._min_prevailing] + self._tp_category
        x_per_row = [self.tp_x_value(t) for t in t_per_row]

        # loop through RH rows and create mesh vertices and faces
        vertices = [Point2D(x, self._base_point.y) for x in x_per_row]
        faces, vert_count, row_len = [], 0, len(t_per_row)
        for to in self._to_category:
            vert_count += row_len
            y1 = self.to_y_value(to)
            vertices.append(Point2D(x_per_row[0], y1))
            for i, t in enumerate(x_per_row[1:]):
                vertices.append(Point2D(x_per_row[i + 1], y1))
                v1 = vert_count - row_len + i
                v2 = v1 + 1
                v3 = vert_count + i + 1
                v4 = v3 - 1
                faces.append((v1, v2, v3, v4))

        # create the Mesh2D, remove unused faces, and assign the colors
        mesh = Mesh2D(vertices, faces)
        mesh = mesh.remove_faces_only(self._remove_pattern)
        mesh.colors = self._container.value_colors
        return mesh

    def _compute_border(self):
        """Compute a Polygon2D for the outer border of the chart."""
        # get properties used to establish the border of the chart
        bpt = self.base_point
        x_max = bpt.x + (self.max_prevailing - self.min_prevailing) * self._x_dim
        y_max = bpt.y + (self.max_operative - self.min_operative) * self._y_dim

        # get the points and build the polyline
        return Polygon2D(
            (bpt, Point2D(x_max, bpt.y), Point2D(x_max, y_max), Point2D(bpt.x, y_max))
        )

    def _process_legend_default(self, l_par):
        """Override the dimensions of the legend to ensure it fits the chart."""
        min_pt, max_pt = self.container.min_point, self.container.max_point
        if l_par.vertical and l_par.is_segment_height_default:
            l_par.properties_3d.segment_height = (max_pt.y - min_pt.y) / 20
            l_par.properties_3d._is_segment_height_default = True
        elif l_par.vertical and l_par.is_segment_height_default:
            l_par.properties_3d.segment_width = (max_pt.x - min_pt.x) / 20
            l_par.properties_3d._is_segment_width_default = True

    @staticmethod
    def _check_number(value, value_name):
        """Check a given value for a dimension input."""
        try:
            value = float(value)
        except (ValueError, TypeError):
            raise TypeError('Expected number for Psychrometric Chart {}. '
                            'Got {}.'.format(value_name, type(value)))
        assert value > 0, 'Psychrometric Chart {} must be greater than 0. ' \
            'Got {}.'.format(value_name, value)
        return value

    def __len__(self):
        """Return length of values on the object."""
        return len(self.operative_temperature._values)

    def __getitem__(self, key):
        """Return a tuple of temperature and humidity."""
        return self.prevailing_outdoor_temperature._values[key], \
            self.operative_temperature._values[key]

    def __iter__(self):
        """Iterate through the values."""
        return zip(
            self.prevailing_outdoor_temperature._values,
            self.operative_temperature._values
        )

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Adaptive Chart representation."""
        return 'Adaptive Chart: {} values'.format(len(self))



# coding=utf-8
"""Utility function for calculating Discomfort Index (DI)."""


def discomfort_index(ta, rh):
    """Calculate discomfort index (DI) from air temperature and relative
    humidity.

    Discomfort Index is derived from original work carried out by Eral C. Thom
    [1] which defined discomfort index based on dry-bulb and wet-bulb
    temperature. It is the human-perceived increase in air temperature due to
    humidity increase.

    Note:
        [1]  Thom, E.C. (1959) "The Discomfort Index". Weatherwise, 12, 57-61.

    Args:
        ta: Air temperature [C]
        rh: Relative humidity  [%]

    Returns:
        di -- Discomfort index [C]
    """
    di = ta - (0.55 - 0.0055 * rh) * (ta - 14.5)

    return di


def discomfort_index_effect_category(di):
    """Get the category of effect associated with a given discomfort index
    (DI).

    Each number (from -6 to 3) represents a certain DI thermal sensation
    category. With categories being the following:

    * -6 = Hyper-glacial
    * -5 = Glacial
    * -4 = Extremely cold
    * -3 = Very cold
    * -2 = Cold
    * -1 = Cool
    * 0 = Comfortable
    * 1 = Hot
    * 2 = Very hot
    * 3 = Torrid

    Args:
        di: Discomfort Index [C]

    Returns:
        category -- An integer indicating the level of effect associated with the
        discomfort index. Values are one of the following:

        -   -6 = Hyper-glacial
        -   -5 = Glacial
        -   -4 = Extremely cold
        -   -3 = Very cold
        -   -2 = Cold
        -   -1 = Cool
        -   0 = Comfortable
        -   1 = Hot
        -   2 = Very hot
        -   3 = Torrid
    """
    if di >= 30:
        category = 3
    elif di >= 26.5:
        category = 2
    elif di >= 20:
        category = 1
    elif di >= 15:
        category = 0
    elif di >= 13:
        category = -1
    elif di >= -1.8:
        category = -2
    elif di >= -10:
        category = -3
    elif di >= -20:
        category = -4
    elif di >= -40:
        category = -5
    else:
        category = -6

    return category



"""Create spatial thermal maps using EnergyPlus and Radiance results."""
import click
import sys
import logging
import json
import os
import shutil

from ladybug.epw import EPW
from ladybug.legend import LegendParameters
from ladybug.color import Colorset
from ladybug.datacollection import HourlyContinuousCollection, \
    HourlyDiscontinuousCollection
from ladybug.header import Header
from ladybug.analysisperiod import AnalysisPeriod
from ladybug.datatype.energyflux import MetabolicRate
from ladybug.datatype.rvalue import ClothingInsulation
from ladybug.datatype.temperature import OperativeTemperature, \
    StandardEffectiveTemperature, Temperature, UniversalThermalClimateIndex
from ladybug.datatype.thermalcondition import PredictedMeanVote, \
    ThermalCondition, ThermalConditionElevenPoint
from ladybug.datatype.temperaturedelta import OperativeTemperatureDelta
from ladybug.datatype.fraction import Fraction

from ladybug_comfort.map.irr import irradiance_contrib_map
from ladybug_comfort.map.mrt import shortwave_mrt_map, longwave_mrt_map
from ladybug_comfort.map.air import air_map
from ladybug_comfort.map.tcp import tcp_model_schedules, tcp_total
from ladybug_comfort.map._enclosure import _parse_enclosure_info, _values_to_data
from ladybug_comfort.collection.pmv import PMV, _PMVnoSET
from ladybug_comfort.collection.adaptive import Adaptive, PrevailingTemperature
from ladybug_comfort.collection.utci import UTCI

from ._helper import load_values, load_analysis_period_str, \
    load_pmv_par_str, load_adaptive_par_str, load_utci_par_str, \
    load_solarcal_par_str, thermal_map_csv, _data_to_ill

_logger = logging.getLogger(__name__)


@click.group(help='Commands for creating spatial thermal maps.')
def map():
    pass


@map.command('pmv')
@click.argument('result-sql', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('enclosure-info', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('epw-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--total-irradiance', '-tr', help='Path to an .ill file output by '
              'Radiance containing total irradiance for each sensor in the '
              'enclosure-info. If unspecified, no shortwave solar will be '
              'assumed for the study.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--direct-irradiance', '-dr', help='Path to an .ill file output by '
              'Radiance containing direct irradiance for each sensor in the '
              'enclosure-info. If unspecified, all shortwave will be assumed '
              'to be indirect.', default=None,
              type=click.Path(file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--ref-irradiance', '-rr', help='Path to an .ill file output by Radiance '
              'containing total ground-reflected irradiance for each sensor in the '
              'enclosure-info. If unspecified, a default ground reflectance of 0.25 '
              'will be assumed for the study.', default=None,
              type=click.Path(file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--sun-up-hours', '-sh', help='Path to a sun-up-hours.txt file output by '
              'Radiance. Required if any irradiance options are provided.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--air-speed', '-v', help='A single number for air speed in m/s or a '
              'string of a JSON array with numbers that align with the result-sql '
              'reporting period. If unspecified, 0.1 m/s will be used.',
              default=None, type=str)
@click.option('--met-rate', '-m', help='A single number for metabolic rate in met '
              'or a string of a JSON array with numbers that align with the '
              'result-sql reporting period. If unspecified, 1.1 met will be used.',
              default=None, type=str)
@click.option('--clo-value', '-c', help='A single number for clothing level in clo '
              'or a string of a JSON array with numbers that align with the '
              'result-sql reporting period. If unspecified, 0.7 clo will be used.',
              default=None, type=str)
@click.option('--run-period', '-rp', help='An AnalysisPeriod string to dictate the '
              'start and end of the analysis (eg. "6/21 to 9/21 between 8 and 16 @1"). '
              'If unspecified, results will be generated for the entire run period of '
              'the result-sql.', default=None, type=str)
@click.option('--write-op-map/--write-set-map', ' /-set', help='Flag to note whether '
              'the output temperature CSV should record Operative Temperature '
              'or Standard Effective Temperature (SET). SET is relatively intense '
              'to compute and so only recording Operative Temperature can greatly '
              'reduce run time, particularly when air speeds are low. However, SET '
              'accounts for all 6 PMV model inputs and so is a more representative '
              '"feels-like" temperature for the PMV model.', default=True)
@click.option('--solarcal-par', '-sp', help='A SolarCalParameter string to customize '
              'the assumptions of the SolarCal model.', default=None, type=str)
@click.option('--comfort-par', '-cp', help='A PMVParameter string to customize the '
              'assumptions of the PMV model.', default=None, type=str)
@click.option('--folder', '-f', help='Folder into which the result CSV files will be '
              'written. If None, files will be written to a "thermal_map" sub-folder in'
              'same directory as the result-sql.', default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--log-file', '-log', help='Optional log file to output the paths to the '
              'generated CSV files. By default this will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def pmv(result_sql, enclosure_info, epw_file,
        total_irradiance, direct_irradiance, ref_irradiance, sun_up_hours,
        air_speed, met_rate, clo_value, write_op_map,
        run_period, comfort_par, solarcal_par, folder, log_file):
    """Get CSV files with maps of PMV comfort from EnergyPlus and Radiance results.

    \b
    Args:
        result_sql: Path to an SQLite file that was generated by EnergyPlus.
            This file must contain hourly or sub-hourly results for zone comfort
            variables.
        enclosure_info: Path to a JSON file containing information about the radiant
            enclosure that sensor points belong to.
        epw_file: Path to an .epw file, used to estimate conditions for any outdoor
            sensors and to provide sun positions.
    """
    try:
        # load the EPW object, run period, air speed, and other parameters
        epw_obj = EPW(epw_file)
        run_period = load_analysis_period_str(run_period)
        air_speed = load_values(air_speed)
        met_rate = load_values(met_rate)
        clo_value = load_values(clo_value)
        solarcal_par = load_solarcal_par_str(solarcal_par)
        comfort_par = load_pmv_par_str(comfort_par)

        # load and align the thermal results from the result_sql file
        pt_air_temps, pt_rad_temps, pt_humids, pt_speeds, a_per = _parse_enclosure_info(
            enclosure_info, result_sql, epw_obj, run_period, air_speed,
            include_humidity=True)

        # adjust the radiant temperature for shortwave solar
        if total_irradiance is not None and os.path.isfile(total_irradiance):
            assert sun_up_hours is not None and os.path.isfile(sun_up_hours), \
                'Sun up hours must be specified when total irradiance is specified.'
            pt_rad_temps = shortwave_mrt_map(
                epw_obj.location, pt_rad_temps, sun_up_hours,
                total_irradiance, direct_irradiance, ref_irradiance,
                solarcal_par=solarcal_par, indirect_is_total=True)

        # convert any input lists of clothing or met to data collections
        met_rate = _values_to_data(met_rate, a_per, MetabolicRate, 'met')
        clo_value = _values_to_data(clo_value, a_per, ClothingInsulation, 'clo')
        if run_period is not None and a_per != run_period:
            met_rate = met_rate.filter_by_analysis_period(run_period) \
                if isinstance(met_rate, HourlyContinuousCollection) else met_rate
            clo_value = clo_value.filter_by_analysis_period(run_period) \
                if isinstance(clo_value, HourlyContinuousCollection) else clo_value

        # run the collections through the PMV model and output results
        comf_class = _PMVnoSET if write_op_map else PMV
        temperature, condition, condition_intensity = [], [], []
        for t_a, rh, t_r, vel in zip(pt_air_temps, pt_humids, pt_rad_temps, pt_speeds):
            pmv_obj = comf_class(
                t_a, rh, t_r, vel, met_rate, clo_value, comfort_parameter=comfort_par)
            condition.append(pmv_obj.thermal_condition)
            condition_intensity.append(pmv_obj.predicted_mean_vote)
            if write_op_map:
                temperature.append(pmv_obj.operative_temperature)
            else:
                temperature.append(pmv_obj.standard_effective_temperature)

        # write out the final results to CSV files
        if folder is None:
            folder = os.path.join(os.path.dirname(result_sql), 'thermal_map')
        result_file_dict = thermal_map_csv(
            folder, temperature, condition, condition_intensity)
        log_file.write(json.dumps(result_file_dict))
    except Exception as e:
        _logger.exception('Failed to run PMV model comfort map.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@map.command('adaptive')
@click.argument('result-sql', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('enclosure-info', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('epw-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--total-irradiance', '-tr', help='Path to an .ill file output by '
              'Radiance containing total irradiance for each sensor in the '
              'enclosure-info. If unspecified, no shortwave solar will be '
              'assumed for the study.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--direct-irradiance', '-dr', help='Path to an .ill file output by '
              'Radiance containing direct irradiance for each sensor in the '
              'enclosure-info. If unspecified, all shortwave will be assumed '
              'to be indirect.', default=None,
              type=click.Path(file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--ref-irradiance', '-rr', help='Path to an .ill file output by Radiance '
              'containing total ground-reflected irradiance for each sensor in the '
              'enclosure-info. If unspecified, a default ground reflectance of 0.25 '
              'will be assumed for the study.', default=None,
              type=click.Path(file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--sun-up-hours', '-sh', help='Path to a sun-up-hours.txt file output by '
              'Radiance. Required if any irradiance options are provided.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--air-speed', '-v', help='A single number for air speed in m/s or a '
              'string of a JSON array with numbers that align with the result-sql '
              'reporting period. If unspecified, 0.1 m/s will be used.',
              default=None, type=str)
@click.option('--run-period', '-rp', help='An AnalysisPeriod string to dictate the '
              'start and end of the analysis (eg. "6/21 to 9/21 between 8 and 16 @1"). '
              'If unspecified, results will be generated for the entire run period of '
              'the result-sql.', default=None, type=str)
@click.option('--solarcal-par', '-sp', help='A SolarCalParameter string to customize '
              'the assumptions of the SolarCal model.', default=None, type=str)
@click.option('--comfort-par', '-cp', help='An AdaptiveParameter string to customize '
              'the assumptions of the Adaptive comfort model.', default=None, type=str)
@click.option('--folder', '-f', help='Folder into which the result CSV files will be '
              'written. If None, files will be written to a "thermal_map" sub-folder in'
              'same directory as the result-sql.', default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--log-file', '-log', help='Optional log file to output the paths to the '
              'generated CSV files. By default this will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def adaptive(result_sql, enclosure_info, epw_file,
             total_irradiance, direct_irradiance, ref_irradiance, sun_up_hours,
             air_speed, run_period, comfort_par, solarcal_par, folder, log_file):
    """Get CSV files with maps of Adaptive comfort from EnergyPlus and Radiance results.

    \b
    Args:
        result_sql: Path to an SQLite file that was generated by EnergyPlus.
            This file must contain hourly or sub-hourly results for zone comfort
            variables.
        enclosure_info: Path to a JSON file containing information about the radiant
            enclosure that sensor points belong to.
        epw_file: Path to an .epw file, used to estimate conditions for any outdoor
            sensors and to provide prevailing outdoor temperature for the adaptive
            comfort model.
    """
    try:
        # load the EPW object, run period, air speed, and other parameters
        epw_obj = EPW(epw_file)
        run_period = load_analysis_period_str(run_period)
        air_speed = load_values(air_speed)
        solarcal_par = load_solarcal_par_str(solarcal_par)
        comfort_par = load_adaptive_par_str(comfort_par)

        # load and align the thermal results from the result_sql file
        pt_air_temps, pt_rad_temps, _, pt_speeds, _ = _parse_enclosure_info(
            enclosure_info, result_sql, epw_obj, run_period, air_speed)

        # adjust the radiant temperature for shortwave solar
        if total_irradiance is not None and os.path.isfile(total_irradiance):
            assert sun_up_hours is not None and os.path.isfile(sun_up_hours), \
                'Sun up hours must be specified when total irradiance is specified.'
            pt_rad_temps = shortwave_mrt_map(
                epw_obj.location, pt_rad_temps, sun_up_hours,
                total_irradiance, direct_irradiance, ref_irradiance,
                solarcal_par=solarcal_par, indirect_is_total=True)

        # compute previaling outdoor temperature so it's not recomputed for each sensor
        avg_month = comfort_par.avg_month_or_running_mean \
            if comfort_par is not None else True
        prev_obj = PrevailingTemperature(epw_obj.dry_bulb_temperature, avg_month)
        prevail_temp = prev_obj.get_aligned_prevailing(pt_air_temps[0])

        # run the collections through the Adaptive model and output results
        temperature, condition, condition_intensity = [], [], []
        for t_air, t_rad, vel in zip(pt_air_temps, pt_rad_temps, pt_speeds):
            adaptive_obj = Adaptive.from_air_and_rad_temp(
                prevail_temp, t_air, t_rad, vel, comfort_parameter=comfort_par)
            temperature.append(adaptive_obj.operative_temperature)
            condition.append(adaptive_obj.thermal_condition)
            condition_intensity.append(adaptive_obj.degrees_from_neutral)

        # write out the final results to CSV files
        if folder is None:
            folder = os.path.join(os.path.dirname(result_sql), 'thermal_map')
        result_file_dict = thermal_map_csv(
            folder, temperature, condition, condition_intensity)
        log_file.write(json.dumps(result_file_dict))
    except Exception as e:
        _logger.exception('Failed to run Adaptive model comfort map.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@map.command('utci')
@click.argument('result-sql', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('enclosure-info', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('epw-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--total-irradiance', '-tr', help='Path to an .ill file output by '
              'Radiance containing total irradiance for each sensor in the '
              'enclosure-info. If unspecified, no shortwave solar will be '
              'assumed for the study.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--direct-irradiance', '-dr', help='Path to an .ill file output by '
              'Radiance containing direct irradiance for each sensor in the '
              'enclosure-info. If unspecified, all shortwave will be assumed '
              'to be indirect.', default=None,
              type=click.Path(file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--ref-irradiance', '-rr', help='Path to an .ill file output by Radiance '
              'containing total ground-reflected irradiance for each sensor in the '
              'enclosure-info. If unspecified, a default ground reflectance of 0.25 '
              'will be assumed for the study.', default=None,
              type=click.Path(file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--sun-up-hours', '-sh', help='Path to a sun-up-hours.txt file output by '
              'Radiance. Required if any irradiance options are provided.', default=None,
              type=click.Path(exists=True, file_okay=True, dir_okay=False,
                              resolve_path=True))
@click.option('--wind-speed', '-v', help='A single number for meteorological wind '
              'speed in m/s or a string of a JSON array with numbers that align with '
              'the result-sql reporting period. This will be used for all indoor '
              'comfort evaluation while the EPW wind speed will be used for the '
              'outdoors. If unspecified, 0.5 m/s will be used.', default=None, type=str)
@click.option('--run-period', '-rp', help='An AnalysisPeriod string to dictate the '
              'start and end of the analysis (eg. "6/21 to 9/21 between 8 and 16 @1"). '
              'If unspecified, results will be generated for the entire run period of '
              'the result-sql.', default=None, type=str)
@click.option('--solarcal-par', '-sp', help='A SolarCalParameter string to customize '
              'the assumptions of the SolarCal model.', default=None, type=str)
@click.option('--comfort-par', '-cp', help='An UTCIParameter string to customize the '
              'assumptions of the Adaptrive comfort model.', default=None, type=str)
@click.option('--folder', '-f', help='Folder into which the result CSV files will be '
              'written. If None, files will be written to a "thermal_map" sub-folder in'
              'same directory as the result-sql.', default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--log-file', '-log', help='Optional log file to output the paths to the '
              'generated CSV files. By default this will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def utci(result_sql, enclosure_info, epw_file,
         total_irradiance, direct_irradiance, ref_irradiance, sun_up_hours,
         wind_speed, run_period, comfort_par, solarcal_par, folder, log_file):
    """Get CSV files with maps of UTCI comfort from EnergyPlus and Radiance results.

    \b
    Args:
        result_sql: Path to an SQLite file that was generated by EnergyPlus.
            This file must contain hourly or sub-hourly results for zone comfort
            variables.
        enclosure_info: Path to a JSON file containing information about the radiant
            enclosure that sensor points belong to.
        epw_file: Path to an .epw file, used to estimate conditions for any outdoor
            sensors and to provide sun positions.
    """
    try:
        # load the EPW object, run period, air speed, and other parameters
        epw_obj = EPW(epw_file)
        run_period = load_analysis_period_str(run_period)
        wind_speed = load_values(wind_speed)
        solarcal_par = load_solarcal_par_str(solarcal_par)
        comfort_par = load_utci_par_str(comfort_par)

        # load and align the thermal results from the result_sql file
        pt_air_temps, pt_rad_temps, pt_humids, pt_speeds, _ = _parse_enclosure_info(
            enclosure_info, result_sql, epw_obj, run_period, wind_speed,
            include_humidity=True, use_10m_wind_speed=True)

        # adjust the radiant temperature for shortwave solar
        if total_irradiance is not None and os.path.isfile(total_irradiance):
            assert sun_up_hours is not None and os.path.isfile(sun_up_hours), \
                'Sun up hours must be specified when total irradiance is specified.'
            pt_rad_temps = shortwave_mrt_map(
                epw_obj.location, pt_rad_temps, sun_up_hours,
                total_irradiance, direct_irradiance, ref_irradiance,
                solarcal_par=solarcal_par, indirect_is_total=True)

        # run the collections through the UTCI model and output results
        temperature, condition, condition_intensity = [], [], []
        for t_a, rh, t_r, vel in zip(pt_air_temps, pt_humids, pt_rad_temps, pt_speeds):
            utci_obj = UTCI(t_a, rh, t_r, vel, comfort_parameter=comfort_par)
            temperature.append(utci_obj.universal_thermal_climate_index)
            condition.append(utci_obj.thermal_condition)
            condition_intensity.append(utci_obj.thermal_condition_eleven_point)

        # write out the final results to CSV files
        if folder is None:
            folder = os.path.join(os.path.dirname(result_sql), 'thermal_map')
        result_file_dict = thermal_map_csv(
            folder, temperature, condition, condition_intensity)
        log_file.write(json.dumps(result_file_dict))
    except Exception as e:
        _logger.exception('Failed to run UTCI model comfort map.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@map.command('irradiance-contrib')
@click.argument('result-sql', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('direct-specular', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('indirect-specular', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('ref-specular', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('indirect-diffuse', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('ref-diffuse', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('sun-up-hours', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--aperture-id', '-id', help='Text string for the identifier of the '
              'aperture associated with the irradiance. If unspecified, it will the '
              'first aperture found in the result-sql, essentially assuming there is '
              'only one dynamic group in the file.', default=None, type=str)
@click.option('--folder', '-f', help='Folder into which the result CSV files will be '
              'written. If None, files will be written to a "contrib" sub-folder in'
              'same directory as the result-sql.', default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--log-file', '-log', help='Optional log file to output the paths to the '
              'generated CSV files. By default this will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def irradiance_contrib(
    result_sql, direct_specular, indirect_specular, ref_specular,
    indirect_diffuse, ref_diffuse, sun_up_hours, aperture_id, folder, log_file
):
    """Get CSV files with irradiance contributions from dynamic windows.

    \b
    Args:
        result_sql: Path to an SQLite file that was generated by EnergyPlus.
            This file must contain results for window transmittance.
        direct_specular: Path to an .ill file output by Radiance containing direct
            irradiance for the specular version of the aperture group.
        indirect_specular: Path to an .ill file output by Radiance containing
            the indirect irradiance for the specular version of the aperture group.
        ref_specular: Path to an .ill file output by Radiance containing ground-reflected
            irradiance for the specular version of the aperture group.
        indirect_diffuse: Path to an .ill file output by Radiance containing
            the indirect irradiance for the diffuse version of the aperture group.
        ref_diffuse: Path to an .ill file output by Radiance containing ground-reflected
            irradiance for the diffuse version of the aperture group.
        sun_up_hours: Path to a sun-up-hours.txt file output by an annual
            irradiance simulation.
    """
    try:
        # compute the irradiance contribution values from the input
        direct_mtx, indirect_mtx, ref_mtx = irradiance_contrib_map(
            result_sql, direct_specular, indirect_specular, ref_specular,
            indirect_diffuse, ref_diffuse, sun_up_hours, aperture_id)

        # prepare the files and directory where the results will be written
        out_folder = folder if folder is not None else \
            os.path.join(os.path.dirname(result_sql), 'contrib')
        if not os.path.isdir(out_folder):
            os.mkdir(out_folder)
        direct_file = os.path.join(out_folder, 'direct.ill')
        indirect_file = os.path.join(out_folder, 'indirect.ill')
        ref_file = os.path.join(out_folder, 'reflected.ill')

        # write the irradiance matrices into CSV files
        _data_to_ill(direct_mtx, direct_file)
        _data_to_ill(indirect_mtx, indirect_file)
        _data_to_ill(ref_mtx, ref_file)
        log_file.write(json.dumps([direct_file, indirect_file, ref_file]))
    except Exception as e:
        _logger.exception('Failed to run Shortwave MRT Delta map.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@map.command('shortwave-mrt')
@click.argument('epw-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('indirect-irradiance', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('direct-irradiance', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('ref-irradiance', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('sun-up-hours', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--contributions', '-c', help='An optional folder containing '
              'sub-folders of irradiance contributions from dynamic aperture groups. '
              'There should be one sub-folder per window groups and each one should '
              'contain three .ill files named direct.ill, indirect.ill and '
              'reflected.ill. If specified, these will be added to the irradiance '
              'inputs before computing shortwave MRT deltas.',
              default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--transmittance-contribs', '-tc', help='An optional folder containing '
              'a transmittance schedule JSON and sub-folders of irradiance results '
              'that exclude the shade from the calculation. There should be one '
              'sub-folder per window groups and each one should contain three .ill '
              'files named direct.ill, indirect.ill and reflected.ill. If specified, '
              'these will be added to the irradiance inputs before computing shortwave '
              'MRT deltas.', default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--trans-schedule-json', '-ts', help='An optional path to a transmittance '
              'schedule JSON output by the honeybee-energy model-transmittance-schedules'
              ' command, which is coordinated with the --transmittance-contribs. '
              'If unspecified, it will be assumed that this JSON already exists in '
              'the root of the --transmittance-contribs with a name schedules.json.',
              default=None, type=click.Path(exists=False, file_okay=True, dir_okay=False,
                                            resolve_path=True))
@click.option('--run-period', '-rp', help='An AnalysisPeriod string to dictate the '
              'start and end of the analysis (eg. "6/21 to 9/21 between 8 and 16 @1"). '
              'If unspecified, results will be annual.', default=None, type=str)
@click.option('--solarcal-par', '-sp', help='A SolarCalParameter string to customize '
              'the assumptions of the SolarCal model.', default=None, type=str)
@click.option('--is-indirect/--indirect-is-total', ' /-t', help='Flag to '
              'note whether the indirect-irradiance argument is actually the total '
              'irradiance, in which case the direct irradiance should be subtracted '
              'from it to get indirect irradiance.', default=True, show_default=True)
@click.option('--output-file', '-f', help='Optional file to output the CSV matrix '
              'of MRT deltas. By default this will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def shortwave_mrt(
        epw_file, indirect_irradiance, direct_irradiance, ref_irradiance,
        sun_up_hours, contributions, transmittance_contribs, trans_schedule_json,
        run_period, solarcal_par, is_indirect, output_file):
    """Get CSV files with maps of shortwave MRT Deltas from Radiance results.

    \b
    Args:
        epw_file: Path to an .epw file, used to estimate conditions for any outdoor
            sensors and to provide sun positions.
        indirect_irradiance: Path to an .ill file output by Radiance containing
            the indirect irradiance for each sensor. Alternatively, if the
            --indirect-is-total input is used, then this input should be the total
            irradiance for each sensor.
        direct_irradiance: Path to an .ill file output by Radiance containing direct
            irradiance for each sensor.
        ref_irradiance: Path to an .ill file output by Radiance containing
            ground-reflected irradiance for each sensor.
        sun_up_hours: Path to a sun-up-hours.txt file output by an annual
            irradiance simulation.
    """
    try:
        # load the EPW object, run period, and other parameters
        epw_obj = EPW(epw_file)
        run_period = load_analysis_period_str(run_period)
        run_period = run_period if run_period is not None else AnalysisPeriod()
        solarcal_par = load_solarcal_par_str(solarcal_par)

        # create a dummy longwave MRT matrix to pass to the shortwave calculator
        header = Header(Temperature(), 'C', run_period)
        lt_values = [0] * len(run_period)
        pt_rad_temps = [HourlyContinuousCollection(header, lt_values)] \
            if run_period.st_hour == 0 and run_period.end_hour == 23 else \
            [HourlyDiscontinuousCollection(header, lt_values, run_period.datetimes)]

        # if the trans_schedule_json is specified, copy it to the contrib folder
        if trans_schedule_json is not None and os.path.isfile(trans_schedule_json):
            if transmittance_contribs is not None and \
                    os.path.isdir(transmittance_contribs):
                sch_json = os.path.join(transmittance_contribs, 'schedules.json')
                shutil.copyfile(trans_schedule_json, sch_json)

        # adjust the radiant temperature for shortwave solar
        is_total = not is_indirect
        d_mrt_temps = shortwave_mrt_map(
            epw_obj.location, pt_rad_temps, sun_up_hours,
            indirect_irradiance, direct_irradiance, ref_irradiance,
            contributions, transmittance_contribs,
            solarcal_par=solarcal_par, indirect_is_total=is_total)

        # write out the final results to CSV files
        if len(d_mrt_temps) == 0:  # no sun-up hours; just create a blank file
            output_file.write('')
        else:
            for mrt_d in d_mrt_temps:
                output_file.write(','.join(str(v) for v in mrt_d))
                output_file.write('\n')
    except Exception as e:
        _logger.exception('Failed to run Shortwave MRT Delta map.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@map.command('longwave-mrt')
@click.argument('result-sql', type=click.Path(
    file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('view-factors', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('modifiers', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('enclosure-info', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('epw-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--run-period', '-rp', help='An AnalysisPeriod string to dictate the '
              'start and end of the analysis (eg. "6/21 to 9/21 between 8 and 16 @1"). '
              'If unspecified, results will be annual.', default=None, type=str)
@click.option('--output-file', '-f', help='Optional file to output the CSV matrix '
              'of longwave MRT. By default this will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def longwave_mrt(result_sql, view_factors, modifiers, enclosure_info, epw_file,
                 run_period, output_file):
    """Get CSV files with maps of longwave MRT from Radiance and EnergyPlus results.

    \b
    Args:
        result_sql: Path to an SQLite file that was generated by EnergyPlus.
            This file must contain hourly or sub-hourly results for zone comfort
            variables.
        view_factors: CSV of spherical view factors to the surfaces in the result-sql.
        modifiers: Path to modifiers file that aligns with the view-factors.
        enclosure_info: Path to a JSON file containing information about the radiant
            enclosure that sensor points belong to.
        epw_file: Path to an .epw file, used to estimate conditions for any outdoor
            sensors and to provide sun positions.
    """
    try:
        # load the run period
        run_period = load_analysis_period_str(run_period)
        run_period = run_period if run_period is not None else AnalysisPeriod()

        # get the longwave MRT data
        mrt_temps = longwave_mrt_map(
            enclosure_info, modifiers, result_sql, view_factors, epw_file, run_period)

        # write out the final results to CSV files
        for mrt_d in mrt_temps:
            output_file.write(','.join(str(v) for v in mrt_d))
            output_file.write('\n')
    except Exception as e:
        _logger.exception('Failed to run Longwave MRT map.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@map.command('air')
@click.argument('result-sql', type=click.Path(
    file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('enclosure-info', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('epw-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--run-period', '-rp', help='An AnalysisPeriod string to dictate the '
              'start and end of the analysis (eg. "6/21 to 9/21 between 8 and 16 @1"). '
              'If unspecified, results will be generated for the entire run period of '
              'the result-sql.', default=None, type=str)
@click.option('--air-temperature/--relative-humidity', ' /-rh', help='Flag to '
              'note whether the the output matrix should be with relative humidity '
              'values instead of air temperature values.',
              default=True, show_default=True)
@click.option('--output-file', '-f', help='Optional file to output the CSV matrix '
              'of values. By default this will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def air_temperature(result_sql, enclosure_info, epw_file,
                    run_period, air_temperature, output_file):
    """Get CSV files with maps of air temperatures or humidity from EnergyPlus results.

    \b
    Args:
        result_sql: Path to an SQLite file that was generated by EnergyPlus.
            This file must contain hourly or sub-hourly results for zone comfort
            variables.
        enclosure_info: Path to a JSON file containing information about the radiant
            enclosure that sensor points belong to.
        epw_file: Path to an .epw file, used to estimate conditions for any outdoor
            sensors and to provide sun positions.
    """
    try:
        # load the run period
        run_period = load_analysis_period_str(run_period)
        run_period = run_period if run_period is not None else AnalysisPeriod()

        # get the air data
        humidity = not air_temperature
        air_data = air_map(enclosure_info, result_sql, epw_file, run_period, humidity)

        # write out the final results to CSV files
        for air_d in air_data:
            output_file.write(','.join(str(v) for v in air_d))
            output_file.write('\n')
    except Exception as e:
        _logger.exception('Failed to run Air Temperature map.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@map.command('map-result-info')
@click.argument('comfort-model', type=str)
@click.option('--run-period', '-rp', help='The AnalysisPeriod string that dictates the '
              'start and end of the analysis (eg. "6/21 to 9/21 between 8 and 16 @1"). '
              'If unspecified, it will be assumed results are for a full year.',
              default=None, type=str)
@click.option('--qualifier', '-q', help='Text for any options used on the comfort '
              'map simulation that change the output data type of results. For example, '
              'the write-set-map text of the pmv map can be passed here to ensure '
              'the output of this command is for SET instead of operative temperature.',
              default=None, type=str, multiple=True)
@click.option('--folder', '-f', help='Result folder into which JSON info files will be '
              'written. If None, the info will only be output to the output-file and '
              'not written into result sub folders.', default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--log-file', '-log', help='Optional file to output the JSON '
              'string of the result info. By default, it will be printed to stdout',
              type=click.File('w'), default='-', show_default=True)
def map_result_info(comfort_model, run_period, qualifier, folder, log_file):
    """Get a JSON that specifies the data type and units for comfort map outputs.

    This JSON is needed by interfaces to correctly parse comfort map results.

    \b
    Args:
        comfort_model: Text for the comfort model of the thermal mapping simulation.
            Choose from: pmv, adaptive, utci.
    """
    try:
        # parse the run period
        run_period = load_analysis_period_str(run_period)
        run_period = run_period if run_period is not None else AnalysisPeriod()

        # get the data type and units from the comfort model
        comfort_model = comfort_model.lower()
        cond, cond_units = ThermalCondition(), 'condition'
        if comfort_model == 'pmv':
            temp, temp_units = OperativeTemperature(), 'C'
            if 'write-set-map' in qualifier:
                temp = StandardEffectiveTemperature()
            cond_i, cond_i_units = PredictedMeanVote(), 'PMV'
        elif comfort_model == 'adaptive':
            temp, temp_units = OperativeTemperature(), 'C'
            cond_i, cond_i_units = OperativeTemperatureDelta(), 'dC'
        elif comfort_model == 'utci':
            temp, temp_units = UniversalThermalClimateIndex(), 'C'
            cond_i, cond_i_units = ThermalConditionElevenPoint(), 'condition'
        else:
            raise ValueError(
                'Comfort model "{}" not recognized. Choose from: {}.'.format(
                    comfort_model, ('pmv', 'adaptive', 'utci')))

        # build up the dictionary of data headers
        temp_header = Header(temp, temp_units, run_period)
        cond_header = Header(cond, cond_units, run_period)
        cond_i_header = Header(cond_i, cond_i_units, run_period)
        result_info_dict = {
            'temperature': temp_header.to_dict(),
            'condition': cond_header.to_dict(),
            'condition_intensity': cond_i_header.to_dict()
        }

        # build up dictionaries of visualization metadata
        tcp_lpar = LegendParameters(colors=Colorset.annual_comfort())
        hsp_lpar = LegendParameters(colors=Colorset.heat_sensation())
        csp_lpar = LegendParameters(colors=Colorset.cold_sensation())
        metric_info_dict = {
            'TCP': {
                'type': 'VisualizationMetaData',
                'data_type': Fraction('Thermal Comfort Percentage').to_dict(),
                'unit': '%',
                'legend_parameters': tcp_lpar.to_dict()
            },
            'HSP': {
                'type': 'VisualizationMetaData',
                'data_type': Fraction('Heat Sensation Percentage').to_dict(),
                'unit': '%',
                'legend_parameters': hsp_lpar.to_dict()
            },
            'CSP': {
                'type': 'VisualizationMetaData',
                'data_type': Fraction('Cold Sensation Percentage').to_dict(),
                'unit': '%',
                'legend_parameters': csp_lpar.to_dict()
            }
        }

        # write the JSON into result sub-folders
        if folder is not None:
            if not os.path.isdir(folder):
                os.makedirs(folder)
            for metric in ('temperature', 'condition', 'condition_intensity'):
                file_path = os.path.join(folder, '{}.json'.format(metric))
                with open(file_path, 'w') as fp:
                    json.dump(result_info_dict[metric], fp, indent=4)
            for metric in ('TCP', 'HSP', 'CSP'):
                file_path = os.path.join(folder, '{}.json'.format(metric))
                with open(file_path, 'w') as fp:
                    json.dump(metric_info_dict[metric], fp, indent=4)

            # write the VTK config file
            config_file = os.path.join(folder, 'config.json')
            cfg = _tcp_config()
            with open(config_file, 'w') as outf:
                json.dump(cfg, outf)

        log_file.write(json.dumps(result_info_dict))
    except Exception as e:
        _logger.exception('Failed to write thermal map info.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@map.command('tcp')
@click.argument('condition-csv', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('enclosure-info', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--schedule', '-s', help='An optional path to a CSV file to specify '
              'the relevant times during which comfort should be evaluated. If '
              'specified, this will override the --occ-schedule-json for both '
              'indoor and outdoor conditions. If both this option and the '
              '--occ-schedule-json are unspecified, it will be assumed that all '
              'times are relevant.',
              default=None, type=click.Path(exists=False, file_okay=True, dir_okay=False,
                                            resolve_path=True))
@click.option('--occ-schedule-json', '-occ', help='Path to an occupancy schedule '
              'JSON output by the honeybee-energy model-occ-schedules command. This '
              'JSON derives the relevant times based on the occupancy schedules of the '
              'energy model and assumes that all outdoor times are relevant. '
              'If both this option and the --schedule are unspecified, all '
              'times of the study will be considered occupied.',
              default=None, type=click.Path(exists=False, file_okay=True, dir_okay=False,
                                            resolve_path=True))
@click.option('--folder', '-f', help='Result folder into which the result CSV files '
              'will be written. If None, they will be output into the same folder as '
              'the condition-csv.', default=None, show_default=True,
              type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option('--log-file', '-log', help='Optional log file to output the paths to the '
              'CSV files. By default, it will be printed to stdout',
              type=click.File('w'), default='-', show_default=True)
def tcp(condition_csv, enclosure_info, schedule, occ_schedule_json, folder, log_file):
    """Compute Thermal Comfort Percent (TCP) from thermal mapping results.

    \b
    Args:
        condition_csv: Path to a CSV file of thermal conditions output by a
            thermal mapping command.
        enclosure_info: Path to a JSON file containing information about the radiant
            enclosure that sensor points belong to.
    """
    try:
        # set the default folder if not specified
        if folder is None:
            folder = os.path.dirname(condition_csv)
        if not os.path.isdir(folder):
            os.makedirs(folder)

        # compute thermal comfort percent
        schedule = schedule if schedule and os.path.isfile(schedule) else None
        if schedule is not None or occ_schedule_json is None:
            tcp_list, hsp_list, csp_list = tcp_total(condition_csv, schedule)
        else:
            tcp_list, hsp_list, csp_list = tcp_model_schedules(
                condition_csv, enclosure_info, occ_schedule_json)

        # write the lists into CSV files
        tcp_file = os.path.join(folder, 'tcp.csv')
        hsp_file = os.path.join(folder, 'hsp.csv')
        csp_file = os.path.join(folder, 'csp.csv')
        with open(tcp_file, 'w') as fp:
            fp.write('\n'.join([str(v) for v in tcp_list]))
            fp.write('\n')
        with open(hsp_file, 'w') as fp:
            fp.write('\n'.join([str(v) for v in hsp_list]))
            fp.write('\n')
        with open(csp_file, 'w') as fp:
            fp.write('\n'.join([str(v) for v in csp_list]))
            fp.write('\n')
        log_file.write(json.dumps([tcp_file, hsp_file, csp_file]))
    except Exception as e:
        _logger.exception('Failed to compute TCP.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


def _tcp_config():
    """Return vtk-config for a thermal comfort map."""
    return {
        "data": [
            {
                "identifier": "Thermal Comfort Percentage",
                "object_type": "grid",
                "unit": "Percentage",
                "path": 'TCP',
                "hide": False,
                "legend_parameters": {
                    "hide_legend": False,
                    "color_set": "annual_comfort",
                    "label_parameters": {
                        "color": [34, 247, 10],
                        "size": 0,
                        "bold": True
                    }
                }
            },
            {
                "identifier": "Heat Sensation Percentage",
                "object_type": "grid",
                "unit": "Percentage",
                "path": 'HSP',
                "hide": False,
                "legend_parameters": {
                    "hide_legend": False,
                    "color_set": "heat_sensation",
                    "label_parameters": {
                        "size": 0,
                        "color": [34, 247, 10],
                        "bold": True
                    }
                }
            },
            {
                "identifier": "Cold Sensation Percentage",
                "object_type": "grid",
                "unit": "Percentage",
                "path": 'CSP',
                "hide": False,
                "legend_parameters": {
                    "hide_legend": False,
                    "color_set": "cold_sensation",
                    "label_parameters": {
                        "color": [34, 247, 10],
                        "size": 0,
                        "bold": True
                    }
                }
            }
        ]
    }
